/// @addtogroup newton
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file newtontst.cpp
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#include <cmath>
#include <stdexcept>
#include <fstream>

#include "interval/doubleInterval.h"
#include "vectalg/vector.h"
#include "vectalg/matrix.h"
#include "vectalg/norm.h"

#include "map/function.h"
#include "map/map.h"
#include "map/c2map.h"

#include "dynset/rect2set.h"

#include "newton/newton.h"
#include "newton/krawczyk.h"

const int DIMENSION=2;
typedef Vector<double,DIMENSION> FloatVector;
typedef Vector<interval,DIMENSION> IntervalVector;
typedef Matrix<double,DIMENSION,DIMENSION> FloatMatrix;
typedef Matrix<interval,DIMENSION,DIMENSION> IntervalMatrix;

typedef capd::map::Map<IntervalMatrix> Map;
typedef capd::map::Map<FloatMatrix> FloatMap;

// For Henon Map we iterate nonrigorous Newton Operator.
void FloatHenonMap()
{

   std::cout << "\n HENON MAP:  H(x,y) = (1 - a*x*x + y, b*x)   where a=1.4,  b=0.3 ";

   FloatMap Henon("var:x,y,a,b;fun:1-a*x*x+y,b*x;");
   Henon.setParameter("a",1.4);
   Henon.setParameter("b",0.3);

   FloatVector x(2);
   x[0]=-1300; x[1] = 1000;

   std::cout << "\n\n Non rigorous NEWTON OPERATOR "
             << "\n Starting point x0 = " << x;

   for(int i=1; i<=5; ++i)
   {
      x = capd::newton::NewtonOperator(x, x, Henon);
      std::cout << "\n x "<< i << " = " << x;
   }
}


// For Henon Map we iterate Krawczyk and Newton Interval Operators.
// Newton Method in this example needs much smaller set then Krawczyk to work.
void HenonMap()
{

   std::cout << "\n\n HENON MAP:  H(x,y) = (1 - a*x*x + y, b*x)   where a=1.4,  b=0.3 ";

   Map Henon("var:x,y,a,b;fun:1-a*x*x+y,b*x;");
   Henon.setParameter("a",interval(1.4));
   Henon.setParameter("b",interval(0.3));

   IntervalVector x0(2), K(2), N(2);


   K[0] = interval(-100,110); K[1]=interval(-100,110);

   std::cout << "\n\n INTERVAL KRAWCZYK OPERATOR "
             << "\n for set X  = " << K;

   for(int i=1; i<=5; ++i)
   {
      x0=midVector(K);
      K = capd::newton::KrawczykOperator(x0, K, Henon);
      std::cout << "\n iteration "<< i << " = " << K;
   }

   N[0]=interval(-.1,.1); N[1] = interval(0.9,1.1);

   std::cout << "\n\n INTERVAL NEWTON OPERATOR "
             << "\n for set X = " << N;

   for(int i=1; i<=5; ++i)
   {
      x0=midVector(N);
      N = capd::newton::NewtonOperator(x0, N, Henon);
      std::cout << "\n iteration "<< i << " = " << N;
   }
}

class MapIterator
{
 public:
  typedef IntervalVector  VectorType;
  typedef IntervalMatrix  MatrixType;
  typedef capd::map::Map<MatrixType> MapType;

  int dim;
  int numberOfIteration;
  MapType map;

  MapIterator(const MapType &f, int iter = 1): map(f)
  {
     numberOfIteration = iter;
     dim = 2;
  }

  // Value of a map - used in Newton for computation in a point
  VectorType operator()(VectorType x)
  {
     for(int i=0; i<numberOfIteration; ++i)
        x = map(x);
     return x;
  }

  // Value and derivative of a Map - used in Krawczyk for computation in a point
  VectorType operator()(VectorType x, MatrixType &dF)
  {
     dF = MatrixType::Identity(dim);
     for(int i=0; i<numberOfIteration; ++i)
     {
        dF = map[x]* dF;
        x = map(x);
     }
     return x;
  }

  // Derivative of a Map - used in both Krawczyk and Newton for computation on a set X
  MatrixType operator[](VectorType x)
  {
     MatrixType dF = MatrixType::Identity(dim);
     for(int i=0; i<numberOfIteration; ++i)
     {
        dF = map[x]* dF;
        x = map(x);
     }
     return dF;
  }

  int getDimension() const
  {
     return dim;
  }

};

// Rigorous proof of existence of zero second iteration of Henon Map
// Instead of class Map we use defined above class MapIterator.
void HenonMap2()
{
    std::cout << "\n\n SECOND ITERATION OF HENON MAP:  \n  H(x,y) = (1 - a*x*x + y, b*x)   where a=1.4,  b=0.3 ";

    Map Henon("var:x,y,a,b;fun:1-a*x*x+y,b*x;");
    Henon.setParameter("a",interval(1.4));
    Henon.setParameter("b",interval(0.3));

    MapIterator map(Henon, 2);

    capd::newton::Krawczyk<MapIterator> henon(map);

    std::cout << "\n\n KRAWCZYK PROOF \n\n ";

    FloatVector x(2);            // Good quess for zero of function
    x[0] = -10./3; x[1]=131./9;
    double size = 1.e-5;

  try{
    capd::newton::KrawczykResult code = henon.proof(x, size);
    std::cout << resultToText(code) ;
    std::cout << "\n\n afer " << henon.numberOfIterations << " iteration of Krawczyk method"
              << "\n set X " << henon.X
              << "\n Krawczyk operator K " << henon.K;
  }
  catch(std::exception& e)
  {
     std::cout << e.what();
  }

}

// -------------------------------------------------------------------------

int main(int, char *[])
{
   //init_fpunit();

   try
   {
      FloatHenonMap();
      HenonMap();
      HenonMap2();
   }catch(std::exception& e)
   {

      std::ofstream plik;
      plik.open("newtonError.txt");
      plik << e.what();
      plik.close();
      return 1;

   }
   return 0;
}


/// @}
