/// @addtogroup auxil
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file commandLineArgs.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_STATISTICS_H_)
#define _STATISTICS_H_

#include <iterator>
#include <functional>
#include <algorithm>
#include <numeric>
#include <vector>
#include <cmath>

template<class Iterator>
typename std::iterator_traits<Iterator>::value_type average(Iterator beg,Iterator end){
  typename std::iterator_traits<Iterator>::value_type result(0);
  result=std::accumulate(beg,end,result);
  return result/std::distance(beg,end);
}

template<typename Scalar>
std::pair<Scalar,Scalar> linearRegression(const std::vector<Scalar>& x,const std::vector<Scalar>& y){
  Scalar xAv=average(x.begin(),x.end());
  Scalar yAv=average(y.begin(),y.end());
  std::vector<Scalar> xBal(x);
  std::vector<Scalar> yBal(y);
  std::transform(xBal.begin(),xBal.end(),xBal.begin(),std::bind2nd(std::minus<Scalar>(),xAv));
  std::transform(yBal.begin(),yBal.end(),yBal.begin(),std::bind2nd(std::minus<Scalar>(),yAv));

  std::pair<Scalar,Scalar> result;
  Scalar s1(0),s2(0);
  result.first=std::inner_product(xBal.begin(),xBal.end(),yBal.begin(),s1)/
               std::inner_product(xBal.begin(),xBal.end(),xBal.begin(),s2);
  result.second=yAv-result.first*xAv;
  return result;
}

template<typename Scalar>
std::pair<Scalar,Scalar> powerRegression(const std::vector<Scalar>& x,const std::vector<Scalar>& y){
  std::vector<Scalar> xLog(x);
  std::vector<Scalar> yLog(y);
  std::transform(xLog.begin(),xLog.end(),xLog.begin(),std::ptr_fun<Scalar,Scalar>(std::log));
  std::transform(yLog.begin(),yLog.end(),yLog.begin(),std::ptr_fun<Scalar,Scalar>(std::log));
  std::pair<Scalar,Scalar> result=linearRegression<Scalar>(xLog,yLog);
  result.second=std::exp(result.second);
  return result;
}

#endif //_STATISTICS_H_

/// @}

