/// @addtogroup covrel
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file hset.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _HSetClass
#define _HSetClass

#include <vector>

namespace capd{
namespace covrel{

/**
   This class is used to store a grid of a h-set in the form
   center[i] + M*r
*/
template<typename MatrixT>
class GridSet
{
public:
   typedef MatrixT MatrixType;
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;
   typedef typename std::vector<VectorType>::iterator iterator;
   typedef typename std::vector<VectorType>::const_iterator const_iterator;
   
   iterator begin()
   {
      return center.begin();
   }
   
   iterator end()
   {
      return center.end();
   }

   const_iterator begin() const
   {
      return center.begin();
   }
   
   const_iterator end() const
   {
      return center.end();
   }

   std::vector<VectorType> center;
   MatrixType C;
   VectorType r;
};

/**
   this is an abstract class for h-sets - from paper by Gidea-Zgliczynski
   'Covering relations ...'  http://www.im.uj.edu.pl/~zgliczyn
*/

template <typename VectorType, typename IVectorType>
class hSet{
public:

   //  methods for nonrigorous computations - used usually in tests and setting up parameters
   virtual bool outside(const VectorType &) const = 0; // verifies if a vector does not intersect support of h-set
   virtual bool inside (const VectorType &) const = 0; // verifies if a vector is inside the support of h-set
   virtual bool across (const VectorType &) const = 0; // verifies if a vector does not intersect N^+
   virtual bool mapaway(const VectorType &) const = 0; // verifies if the projection onto stable directions
                                                       // of a vector is outside the unit ball in a proper
                                                       // coordinate system

   // interval (rigorous) versions:
   virtual bool outside(const IVectorType &) const = 0;
   virtual bool inside (const IVectorType &) const = 0;
   virtual bool across (const IVectorType &) const = 0;
   virtual bool mapaway(const IVectorType &) const = 0;

   virtual ~hSet(){}
}; //the end of abstract class hSet

}} // namespace capd::covrel

#endif // _HSetClass

/// @}
