/// @addtogroup covrel
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file hsetND.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _HSetND_HPP
#define _HSetND_HPP

#include <stdexcept>
#include "capd/minmax.h"

namespace capd{
namespace covrel{

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::hSetND(
      const VectorType& A_C,
      const MatrixType& A_M,
      int A_d1, int A_d2
   ): m_C(A_C),
      m_M(A_M),
      m_IC(m_C),
      m_d1(A_d1),
      m_d2(A_d2)
{
   if( A_C.dimension()!=A_M.numberOfRows() ||  A_C.dimension()!=A_M.numberOfColumns() )
   {
      throw std::runtime_error("hSetND constructor error: incorrect dimensions of vector or matrix");
   }
   if( m_d1+m_d2!=m_C.dimension() || m_d1<0 || m_d2<0 )
   {
      throw std::runtime_error("hSetND constructor error: incorrect number of unstable and stable directions");
   }
   int i,j;
   for(i=0;i<m_C.dimension();i++)
      for(j=0;j<m_C.dimension();j++)
         m_IM[i][j] = m_M[i][j];

   m_invM = inverseMatrix(m_M);
   m_invIM = inverseMatrix(m_IM);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::hSetND(
            const IVectorType& A_C,
            const IMatrixType& A_M,
            int A_d1, int A_d2)
   : m_IC(A_C), m_IM(A_M),
     m_d1(A_d1), m_d2(A_d2)
{
   if( A_C.dimension()!=A_M.numberOfRows() || A_C.dimension()!=A_M.numberOfColumns() )
   {
      throw std::runtime_error("hSetND constructor error: incorrect dimensions of vector or matrix");
   }
   if( m_d1+m_d2!=m_C.dimension() || m_d1<0 || m_d2<0 )
   {
      throw std::runtime_error("hSetND constructor error: incorrect number of unstable and stable directions");
   }

   int i,j;
   for(i=0;i<m_C.dimension();i++)
   {
      m_C[i] = m_IC[i].mid().leftBound();
      for(j=0;j<m_C.dimension();j++)
      {
         m_M[i][j] = m_IM[i][j].mid().leftBound();
      }
   }

   m_invM = inverseMatrix(m_M);
   m_invIM = inverseMatrix(m_IM);
}

// ------------------- float version ---------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::outside(const VectorType& point) const
{
   VectorType x = m_invM*(point-m_C);
   // we verify if x is outside unit ball in maximum norm
   int i=x.dimension();
   do{
      i--;
      if(capd::abs(x[i])>1)
         return true;
   }while(i);
   return false;
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::inside(const VectorType& point) const
{
   VectorType x = m_invM*(point-m_C);
   // we verify if x is inside unit ball in maximum norm
   int i=x.dimension();
   do{
      i--;
      if(capd::abs(x[i])>=1)
         return false;
   }while(i);
   return true;
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::across(const VectorType& point) const
{
   // we verify if x is inside a horizontal strip, i.e.
   //   x\in B_n (open ball)
   // or
   //   x projected onto unstable directions is outside unit ball
   // notice, that d1 or d2 may be equal to zero

   VectorType x = m_invM*(point-m_C);

   // return true if any unstable coordinate is greater than 1
   int i=0;
   while(i!=m_d1)
   {
      if(capd::abs(x[i])>1)
         return true;
      i++;
   }
   // else all stable coordinates must be lower than 1
   while(i!=x.dimension())
   {
      if(capd::abs(x[i])>=1)
         return false;
      i++;
   }
   return true;
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::mapaway(const VectorType& point) const
{
   // we verify if x projected onto unstable directions is outside unit ball
   // notice, that d1 or d2 may be equal to zero
   if(m_d1==0)
      return true;

   VectorType x = m_invM*(point-m_C);
   int i=0;
   while(i!=m_d1)
   {
      if(capd::abs(x[i])>1)
         return true;
      i++;
   }
   return false;
}


// ------------------------ interval version ---------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::outside(const IVectorType& point) const
{
   IVectorType x = m_invIM*(point-m_IC);
   return checkOutside(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::inside(const IVectorType& point) const
{
   IVectorType x = m_invIM*(point-m_IC);
   return checkInside(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::across(const IVectorType& point) const
{
   IVectorType x = m_invIM*(point-m_IC);
   return checkAcross(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::mapaway(const IVectorType& point) const
{
   IVectorType x = m_invIM*(point-m_IC);
   return checkMapaway(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::outside(
      const capd::dynset::C0set<IMatrixType>& A_set) const
{
   IVectorType x = A_set.affineTransformation(m_invIM,m_IC);
   return checkOutside(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::inside(
      const capd::dynset::C0set<IMatrixType>& A_set) const
{
   IVectorType x = A_set.affineTransformation(m_invIM,m_IC);
   return checkInside(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::across(
      const capd::dynset::C0set<IMatrixType>& A_set) const
{
   IVectorType x = A_set.affineTransformation(m_invIM,m_IC);
   return checkAcross(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::mapaway(
      const capd::dynset::C0set<IMatrixType>& A_set) const
{
   IVectorType x = A_set.affineTransformation(m_invIM,m_IC);
   return checkMapaway(x);
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::checkOutside(const IVectorType& x) const
{
   // we verify if x is outside unit ball in maximum norm
   int i=x.dimension();
   do{
      i--;
      if(capd::abs(x[i])>1)
         return true;
   }while(i);
   return false;
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::checkInside(const IVectorType& x) const
{
   // we verify if x is inside unit ball in maximum norm
   int i=x.dimension();
   do{
      i--;
      if(capd::abs(x[i])>=1)
         return false;
   }while(i);
   return true;
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::checkAcross(const IVectorType& x) const
{
   // we verify if x is inside a horizontal strip, i.e.
   //   x\in B_n (open ball)
   // or
   //   x projected onto unstable directions is outside unit ball
   // notice, that d1 or d2 may be equal to zero

   // return true if any unstable coordinate is greater than 1
   int i=0;
   while(i!=m_d1)
   {
      if(capd::abs(x[i])>1)
         return true;
      i++;
   }
   // else all stable coordinates must be lower than 1
   while(i!=x.dimension())
   {
      if(capd::abs(x[i])>=1)
         return false;
      i++;
   }
   return true;
}

// ---------------------------------------------------------------

template<typename VectorType, typename IVectorType, typename MatrixType, typename IMatrixType>
bool hSetND<VectorType,IVectorType,MatrixType,IMatrixType>::checkMapaway(const IVectorType& x) const
{
   // we verify if x projected onto unstable directions is outside unit ball
   // notice, that d1 or d2 may be equal to zero
   if(m_d1==0)
      return true;

   int i=0;
   while(i!=m_d1)
   {
      if(capd::abs(x[i])>1)
         return true;
      i++;
   }
   return false;
}

// ---------------------------------------------------------------

}} //namespace capd::covrel

#endif // _HSetND_HPP

/// @}
