/// @addtogroup diffIncl
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file inclRec2SetIntf.h
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2007 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _InclRect2SetIntf_H_
#define _InclRect2SetIntf_H_

//#include "dynset/inclc0set.h"
#include "dynset/rect2set.h"


namespace capd{
namespace diffIncl{

///////////////////////////////////////////////////////////////////////////
// InclRect2Set
/// 
/// Set representation for differential inclusions based on capd::dynset::Rect2Set class
///
///   set is represented as: x + C*r0 + B*r   where
///       C*r0 - basic 'Lipschitz part'
///       B*r - QR-decomposition of the remaining errors
///   

template<typename MatrixT>
class InclRect2Set : public capd::dynset::Rect2Set<MatrixT>{

public:
  typedef MatrixT MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef Norm<VectorType,MatrixType> NormType;

  // constr
  explicit InclRect2Set(int, double s_factor = 0.);
  explicit InclRect2Set(const VectorType& the_x, double s_factor = 0.);
  InclRect2Set(const VectorType& the_x, const VectorType& the_r0, double s_factor = 0.);
  InclRect2Set(const VectorType& the_x, const MatrixType& the_C, const VectorType& the_r0, double s_factor = 0.);
  InclRect2Set(const VectorType& the_x, const MatrixType& the_C,
               const VectorType& the_r0,const MatrixType& the_B,
               const VectorType& the_r, double s_factor = 0.
  );
  
  capd::dynset::C0set<MatrixType>* clone(void);
  capd::dynset::C0set<MatrixType>* fresh(void);
  capd::dynset::C0set<MatrixType>* cover(const VectorType& v);
  
  void move(capd::dynsys::DynSys<MatrixType>& dynsys);
  void move(capd::dynsys::DynSys<MatrixType>& dynsys, InclRect2Set& result) const;
  
  std::vector<VectorType> getCorners() const;

  using capd::dynset::Rect2Set<MatrixT>::get_x;
  using capd::dynset::Rect2Set<MatrixT>::get_r;
  using capd::dynset::Rect2Set<MatrixT>::get_r0;
  using capd::dynset::Rect2Set<MatrixT>::get_B;
  using capd::dynset::Rect2Set<MatrixT>::get_C;
  using capd::dynset::Rect2Set<MatrixT>::size;
  using capd::dynset::Rect2Set<MatrixT>::center;
  using capd::dynset::Rect2Set<MatrixT>::operator VectorType;
  using capd::dynset::Rect2Set<MatrixT>::show;
  using capd::dynset::Rect2Set<MatrixT>::affineTransformation;

protected:
  using capd::dynset::Rect2Set<MatrixT>::m_x;
  using capd::dynset::Rect2Set<MatrixT>::m_r;
  using capd::dynset::Rect2Set<MatrixT>::m_r0;
  using capd::dynset::Rect2Set<MatrixT>::m_B;
  using capd::dynset::Rect2Set<MatrixT>::m_C;
};

// inline definitions

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(int dim, double s_factor)
  :  capd::dynset::Rect2Set<MatrixType>(dim, s_factor) {
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(const VectorType& the_x, double s_factor)
  :  capd::dynset::Rect2Set<MatrixType>(the_x, s_factor) {
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(const VectorType& the_x,const VectorType& the_r0, double s_factor)
  :  capd::dynset::Rect2Set<MatrixType>(the_x, the_r0, s_factor) {
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(
      const VectorType& the_x,
      const MatrixType& the_C,
      const VectorType& the_r0, 
      double s_factor
   )
  : capd::dynset::Rect2Set<MatrixType>(the_x, the_C, the_r0, s_factor){
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(
      const VectorType& the_x,
      const MatrixType &the_C,
      const VectorType& the_r0,
      const MatrixType &the_B,
      const VectorType& the_r,
      double s_factor
   ): capd::dynset::Rect2Set<MatrixType>(the_x, the_C, the_r0, the_B, the_r, s_factor){
}

template<typename MatrixType>
inline typename capd::dynset::C0set<MatrixType>* InclRect2Set<MatrixType>::clone(){
   return new InclRect2Set(*this);
}

template<typename MatrixType>
inline typename capd::dynset::C0set<MatrixType>* InclRect2Set<MatrixType>::fresh(){
   return new InclRect2Set<MatrixType>(m_x.dimension());
}

template<typename MatrixType>
inline typename capd::dynset::C0set<MatrixType>* InclRect2Set<MatrixType>::cover(const VectorType& v){
   return new InclRect2Set(midVector(v),v-midVector(v));
}

}} // namespace capd::diffIncl

#endif // _InclRect2SetIntf_H

/// @}
