/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file basicTaylorIntf.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2001-2005 */

#ifndef _BasicTaylorClass
#define _BasicTaylorClass

#include <string>
#include "map/mapintf.h"

namespace capd{
namespace dynsys{

template<typename MapT>
class BasicTaylor
{
public:
   typedef MapT MapType;
   typedef typename MapT::FunctionType FunctionType;
   typedef typename MapType::MatrixType MatrixType;
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;

   BasicTaylor(const MapType& Field, int Order, const ScalarType& Step);
   BasicTaylor(const std::string& Field, int Order, const ScalarType& Step);
   BasicTaylor(const char* Field, int Order, const ScalarType& Step);
   BasicTaylor(const BasicTaylor &);
   virtual ~BasicTaylor();

   BasicTaylor& operator=(const BasicTaylor&);
   VectorType operator()(const VectorType &) const;
   VectorType operator()(const VectorType &, MatrixType &derivative) const;

   const MapType& getField() const;
   virtual void setField(const MapType& Field);

   int getOrder() const;
   virtual void setOrder(int newOrder);

   const ScalarType& getStep() const;
   void setStep(const ScalarType& newStep);

   VectorType *getCoeff();
   MatrixType *getMatrixCoeff();

   void setParameter(const std::string &name, const ScalarType& value);

protected:
   void computeMatrixCoeff(int, const MatrixType& H) const;
   MapType m_vField;

   int m_order;
   ScalarType m_step;
   VectorType *m_coeff;
   MatrixType *m_matrixCoeff;
   int m_dimension;

   MatrixType *m_F; // temp table used in the computations of variational part
};

// -----------------------------------------------------------------------------

template<typename MapType>
inline const MapType& BasicTaylor<MapType>::getField() const
{
   return m_vField;
}

template<typename MapType>
inline int BasicTaylor<MapType>::getOrder() const
{
   return m_order;
}

template<typename MapType>
inline const typename BasicTaylor<MapType>::ScalarType& BasicTaylor<MapType>::getStep() const
{
   return m_step;
}

template<typename MapType>
inline void BasicTaylor<MapType>::setStep(const ScalarType& newStep)
{
   m_step = newStep;
}

template<typename MapType>
inline typename BasicTaylor<MapType>::VectorType* BasicTaylor<MapType>::getCoeff()
{
   return m_coeff;
}

template<typename MapType>
inline typename BasicTaylor<MapType>::MatrixType* BasicTaylor<MapType>::getMatrixCoeff()
{
   return m_matrixCoeff;
}

template<typename MapType>
inline void BasicTaylor<MapType>::setParameter(const std::string& name, const ScalarType& value)
{
   m_vField.setParameter(name, value);
}

}} // namespace capd::dynsys

#endif // define _BasicTaylorClass

/// @}
