/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file funccons.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/*
   Daniel Wilczak, October 2000 - 2004
   This file contains the implementation of constructors, destructor
*/

#ifndef _FUNCCONS_HPP
#define _FUNCCONS_HPP

#include <stdexcept>
#include "map/funcintf.h"
#include "map/textop.h"

namespace capd{
namespace map{

/*______________________ createDefault ________________________________*/

template<typename VectorType>
void Function<VectorType>::createDefault()
{
   m_diff = new std::string[1];
   m_fun = "0";
   m_diff[0] = "?";
   m_tree = new ConsNode<ScalarType>(ScalarType(0),m_order);
}

/*________________________ deleteTables _______________________________*/
template<typename VectorType>
void Function<VectorType>::deleteTables()
{
   delete []m_diff;
   m_diff = 0;
   delete m_tree;
   m_tree = 0; // NULL
}


/*__________________________ CONSTRUCTOR  _____________________________*/

/* This constructor creates a constant function  equal to 0
    with one indefinite variable
*/

template<typename VectorType>
Function<VectorType>::Function(void) 
{
   createDefault();
}

/*__________________________ CONSTRUCTOR  _____________________________*/

/* This constructor reads variables and formula from text parameter */

template<typename VectorType>
Function<VectorType>::Function(const char *s, int i) : BasicFunction<ScalarType>(s,i)
{
   try{
      createFromText(s,i);
   } catch(std::runtime_error &r)
   {
      std::string re = "exception in constructor Function(const char *, int)\n";
      re += r.what();
      throw std::runtime_error(re);
   }
}

/*________________________ COPY CONSTRUCTOR  __________________________*/

template<typename VectorType>
Function<VectorType>::Function(const Function &f) : BasicFunction<ScalarType>(f)
{
   try{
      createFromObject(f);
   } catch(std::runtime_error &r)
   {
      std::string re = "exception in constructor Function(const Function &)\n";
      re += r.what();
      throw std::runtime_error(re);
   }
}

/*__________________________ CONSTRUCTOR  _____________________________*/

template<typename VectorType>
Function<VectorType>::Function(const std::string &f, int i) : BasicFunction<ScalarType>(f,i)
{
   try{
      createFromText(f,i);
   } catch(std::runtime_error &r)
   {
      std::string re = "exception in constructor Function(const string &, int)\n";
      re += r.what();
      throw std::runtime_error(re);
   }
}

/*__________________________ DESTRUCTOR _____________________________*/

/* destructor deletes the dynamic variables: m_tree, stack of the constants
   and possibly stack of IntervalType constants
*/

template<typename VectorType>
Function<VectorType>::~Function(void)
{
   deleteTables();
}

// -----------------------------------------------------------------------------

template<typename VectorType>
void Function<VectorType>::createFromText(const std::string &s, int a_order)
{
   size_t start = s.find(std::string("fun:"));
   size_t last = s.find_first_of(';',start);
   if(last==std::string::npos)
   {
      std::string message("Cannot find delimiter ';' in the formula ");
      throw std::runtime_error(message + s);
   }

   m_fun = s.substr(start+4,last-start-4);  /* we read the formula */
   eqnanal(m_fun,&m_tree);

   m_diff = new std::string[m_dim];
   for(int k=0;k<m_dim;++k)
      m_diff[k] = "?";
}

// -----------------------------------------------------------------------------

template<typename VectorType>
void Function<VectorType>::createFromObject(const Function &f)
{
   m_fun = f.m_fun;
   m_diff = new std::string[m_dim];

   for(int i=0;i<m_dim;++i)
      m_diff[i] = f.m_diff[i];

   eqnanal(m_fun,&m_tree);
}

}} // namespace capd::map

#endif // define _FUNCCONS_HPP

/// @}
