/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file muldiv.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _MULDIV_HPP
#define _MULDIV_HPP

namespace capd{
namespace map{

// ------------------------------- contructors -------------------------------------

template<typename ScalarType>
MulNode<ScalarType>::MulNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

template<typename ScalarType>
DivNode<ScalarType>::DivNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

template<typename ScalarType>
MulConsNode<ScalarType>::MulConsNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

template<typename ScalarType>
DivConsNode<ScalarType>::DivConsNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

// -------------------------------- destructors -----------------------------------

template<typename ScalarType>
MulNode<ScalarType>::~MulNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

template<typename ScalarType>
DivNode<ScalarType>::~DivNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

template<typename ScalarType>
MulConsNode<ScalarType>::~MulConsNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

template<typename ScalarType>
DivConsNode<ScalarType>::~DivConsNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

// ------------------------------ operators ----------------------------------

template<typename ScalarType>
ScalarType& MulNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   (*left)(i);
   (*right)(i);

   ScalarType result(0);
   for(int j=0;j<=i;++j)
      result += (left->value[j]) * (right->value[i-j]) ;

   return value[i] = result;
}

template<typename ScalarType>
ScalarType& DivNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   (*left)(i);
   (*right)(i);
   if (! (right->value[0]!=ScalarType(0.) ) )
   {
      throw(std::runtime_error("function error: division by zero"));
   }

   ScalarType result(0.);
   for(int j=0;j<i;++j)
      result+=value[j] * (right->value[i-j]);

   return value[i] = (left->value[i] - result) / (right->value[0]);
}

template<typename ScalarType>
ScalarType& MulConsNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   bool isZero=true;
   for(int j=0;j<left->getOrder();++j)
      if(left->value[j]!=ScalarType(0))
      {
         isZero=false;
         break;
      }
   if(isZero)
      return value[i] = ScalarType(0);

   (*right)(i);
   ScalarType result(0);
   for(int j=0;j<=i;++j)
      result += (left->value[j]) * (right->value[i-j]) ;
   return value[i]=result;
}

template<typename ScalarType>
ScalarType& DivConsNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   (*left)(i);

   return value[i] = left->value[i] / right->value[0];
}

// ------------------------------- setOrder ----------------------------------

template<typename ScalarType>
void MulNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

template<typename ScalarType>
void DivNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

template<typename ScalarType>
void MulConsNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

template<typename ScalarType>
void DivConsNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

}} // namespace capd::map

// --------------------------------- global Operators -------------------------

template<typename ScalarType>
capd::map::Node<ScalarType>& operator*(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y)
{
   if(x.getOrder()!=y.getOrder())
      throw std::runtime_error("operator*(Node&, Node&) - incompatible dimensions");

   if (x.isConst())
      return *(new capd::map::MulConsNode<ScalarType>(&x,&y,x.getOrder()));
   if (y.isConst())
      return *(new capd::map::MulConsNode<ScalarType>(&y,&x,x.getOrder()));

   return *(new capd::map::MulNode<ScalarType>(&x,&y,x.getOrder()));
}

template<typename ScalarType>
capd::map::Node<ScalarType>& operator/(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y)
{
   if(x.getOrder()!=y.getOrder())
      throw std::runtime_error("operator/(Node&, Node&) - incompatible dimensions");
   return *(new capd::map::DivNode<ScalarType>(&x,&y,x.getOrder()));   
}


#endif // define _MULDIV_HPP

/// @}
