/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file nodeintf.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2004-2005 */

#ifndef _NODEINTF_H
#define _NODEINTF_H

#include <ostream>

namespace capd{
namespace map{

template<typename ScalarType>
class Node
{
protected:
   int m_maxComputedDerivative;
   int m_order;
   Node(int a_order, Node* a_left, Node* a_right)
      : m_maxComputedDerivative(-1), m_order(a_order), m_links(0), left(a_left), right(a_right)
   {}
public:
   void reset();
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;
   int m_links;
   Node *left,*right;                     /* left and right leaf of the tree*/
   ScalarType *value;                     /* value of leaf */
   virtual ScalarType& operator()(int) = 0;  /* operator or function (+,-,*,/,^,sin,cos,exp,log) */
   virtual void setOrder(int, ScalarType*) = 0;
   virtual bool isConst() const {return false;} 
   int getOrder() const { return m_order; }
   void computeDerivatives(int a_upTo);
   
   
   ScalarType& operator[](int i)
   {
      return *(value+i);
   }
   
   const ScalarType& operator[](int i) const
   {
      return *(value+i);
   }

   iterator begin()
   {
      return value;
   }
   iterator end()
   {
      return value+m_order;
   }
   
   const_iterator begin() const
   {
      return value;
   }
   const_iterator end() const
   {
      return value+m_order;
   }
   virtual ~Node();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class ConsNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;
   
   ScalarType& operator()(int i)
   {
      return *(value+i);
   }
   void setOrder(int, ScalarType*);
   bool isConst() const {return true;}
   ConsNode(ScalarType d, int m_order);
   ~ConsNode()
   {
      delete []value;
   }
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class VarNode : public Node<ScalarType>
{
private:
   int m_numberOfVariables;
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int i)
   {
      return *(value+i);
   }
   void setOrder(int, ScalarType*);

   VarNode(ScalarType *d, int m_order, int variable);
   ~VarNode(){}
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class SinNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   SinNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~SinNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class CosNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   CosNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~CosNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class ExpNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   ExpNode(Node<ScalarType> *L, int m_order);
   ~ExpNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class LogNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   LogNode(Node<ScalarType> *L, int m_order);
   ~LogNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class PowNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   PowNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~PowNode();
};

template <>
double& PowNode<double>::operator()(int);

// -----------------------------------------------------------------------------

template<typename ScalarType>
class SqrtNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   SqrtNode(Node<ScalarType> *L, int m_order);
   ~SqrtNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class SumNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   SumNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~SumNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class DifNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   DifNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~DifNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class MulNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   MulNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~MulNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class DivNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;

   ScalarType& operator()(int);
   void setOrder(int, ScalarType*);

   DivNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~DivNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class MulConsNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;
   
   ScalarType& operator()(int i);
   void setOrder(int, ScalarType*);
   MulConsNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~MulConsNode();
};

// -----------------------------------------------------------------------------

template<typename ScalarType>
class DivConsNode : public Node<ScalarType>
{
   using Node<ScalarType>::m_order;
   using Node<ScalarType>::m_maxComputedDerivative;   
public:
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   using Node<ScalarType>::m_links;
   using Node<ScalarType>::left;
   using Node<ScalarType>::right;
   using Node<ScalarType>::value;
   using Node<ScalarType>::begin;
   using Node<ScalarType>::end;
   
   ScalarType& operator()(int i);
   void setOrder(int, ScalarType*);
   DivConsNode(Node<ScalarType> *L, Node<ScalarType> *R, int m_order);
   ~DivConsNode();
};

}} // the end of the namespace capd::map

// -----------------------------------------------------------------------------

template<typename ScalarType>
capd::map::Node<ScalarType>& operator+(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y);

template<typename ScalarType>
capd::map::Node<ScalarType>& operator-(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y);

template<typename ScalarType>
capd::map::Node<ScalarType>& operator*(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y);

template<typename ScalarType>
capd::map::Node<ScalarType>& operator/(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y);

template<typename ScalarType>
std::ostream& operator<<(std::ostream&, const capd::map::Node<ScalarType>&);

#endif // #define _NODEINTF_H

/// @}
