/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file sincos.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _SINCOS_HPP
#define _SINCOS_HPP

#include <stdexcept>
#include <sstream>
#include <cmath>
#include "map/nodeintf.h"

namespace capd{
namespace map{

//  ----------------------- constructors --------------------------------

template<typename ScalarType>
SinNode<ScalarType>::SinNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

template<typename ScalarType>
CosNode<ScalarType>::CosNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

// -------------------------- destructors -------------------------------

template<typename ScalarType>
SinNode<ScalarType>::~SinNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

template<typename ScalarType>
CosNode<ScalarType>::~CosNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

// -------------------------- operators ----------------------------------

template<typename ScalarType>
ScalarType& SinNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   (*left)(i);
   if(i==0)
   {
      right->value[i] = cos(left->value[0]);
      return value[i] = sin(left->value[0]);
   }
   ScalarType result(0.);
   for(int j=0;j<i;++j)
      result += (i-j) * value[j] * left->value[i-j];
   right->value[i] = -result/ScalarType(i);

   result=ScalarType(0.);
   for(int k=0;k<i;++k)
      result += (i-k) * (right->value[k]) * (left->value[i-k]);

   return value[i] = result/ScalarType(i);
}

template<typename ScalarType>
ScalarType& CosNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   (*left)(i);
   if(i==0)
   {
      right->value[i] = sin(left->value[0]);
      return value[i] = cos(left->value[0]);
   }

   ScalarType result(0.);
   for(int j=0;j<i;++j)
      result += (i-j) * value[j] * left->value[i-j];
   right->value[i] = result/ScalarType(i);

   result=ScalarType(0.);
   for(int k=0;k<i;++k)
      result += (i-k) * right->value[k] * left->value[i-k];

   return value[i] = -result/ScalarType(i);
}

// ---------------------- setOrder ---------------------------------

template<typename ScalarType>
void SinNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

template<typename ScalarType>
void CosNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];
   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

}} // namespace capd::map

#endif // define _SINCOS_HPP

/// @}
