/////////////////////////////////////////////////////////////////////////////
//   Package:          CAPD
//   Subpackage:
/// @addtogroup intervals
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file mpInterval.cpp
///
/// @author Tomasz Kapela   @date 23-08-2006
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Tomasz Kapela 2006 
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.


#ifndef _MP_INTERVAL_CPP_
#define _MP_INTERVAL_CPP_

#include <exception>
#include "multiPrec/mpfrClass.h"
#include "interval/interval.h"
#include "interval/mpIntervalSettings.h"
#include "interval/mpIntervalFun.h"
#include <fstream>

#ifdef __HAVE_MPFR__

namespace capd{
namespace intervals{



//////////////////////////////////////////////////////////////////////////
//  quadrant
///
/// returns the integer part of the division of x by Pi/2        
/// the result is exact                                                  
//////////////////////////////////////////////////////////////////////////
Interval<MpfrClass>::BoundType quadrant(const Interval<MpfrClass>::BoundType &x)
{
  int ok = 0;
  capd::multiPrec::PrecisionType defPrec = x.getDefaultPrecision(), 
                                 prec = x.getPrecision(),
                                 precIncreaseRate = 64;
  Interval<MpfrClass>  twoOverPi, tmp;
  Interval<MpfrClass>::BoundType quad = 0.0;
            
  if(!(x==0))
  {
     do{
        x.setDefaultPrecision(prec);
        twoOverPi = Interval<MpfrClass>(2.0) / Interval<MpfrClass>::pi();
        tmp = twoOverPi * x;
        ok = compare(floor(tmp.leftBound()), floor(tmp.rightBound()));

        if( ok != 0 )
           prec += precIncreaseRate;
           
     }while(ok != 0);
     
     quad = floor(tmp.leftBound());
     Interval<MpfrClass>::BoundType::setDefaultPrecision(defPrec);
  }
  return quad;
}

int modulo4(MpfrClass x)
{
  while( x > 32000)
    x =  x - MpfrClass(64000);
  while(x < -32000)
    x = x + MpfrClass(64000);
  
  return ( toInt(x) % 4 );
}

template<>
Interval<MpfrClass> sin(const Interval<MpfrClass>& x)
{
  Interval<MpfrClass>::BoundType quad_left, quad_right;
  Interval<MpfrClass>::BoundType left, right;
  
  int ql_mod4, qr_mod4;

  testNaN(x);

#ifdef _MPI_TEST_INF_  
  if (isinf(x)) {
    /* the two endpoints are the same infinite */
    if ( compare(x.leftBound(), x.rightBound()) == 0) 
       throw NaNException();
   
    return Interval<MpfrClass>(-1.0, 1.0);
  }
#endif  //_MPI_TEST_INF_ 
  

  // quad_left gives the quadrant where the left endpoint of b is 
  // quad_left = floor(2 b->left / pi)                            
  // idem for quad_right and b->right                             
 
  quad_left = quadrant(x.leftBound());
  quad_right = quadrant(x.rightBound());


  if (quad_right - quad_left >= 4.0) 
  {
     return Interval<MpfrClass> (-1, 1);
  } else {
  // there is less than one period in b 
  // let us discuss according to the position (quadrant) of the endpoints of b    
 
    // qr_mod4 gives the quadrant where the right endpoint of b is 
    // qr_mod4 = floor(2 b->right / pi) mod 4 
    qr_mod4 = modulo4(quad_right);
  
    // quad_left gives the quadrant where the left endpoint of b is 
    // quad_left = floor(2 b->left / pi) mod 4 
    ql_mod4 = modulo4(quad_left);  

    switch (qr_mod4) {
      case 0:
        switch (ql_mod4) {
          case 0:
          case 3:
            MpfrClass::roundDown();
            left = sin(x.leftBound());
            MpfrClass::roundUp();
            right = sin(x.rightBound());  
            break;
                    
          case 1:
            MpfrClass::roundUp();
            left = sin(x.leftBound());
            right = sin(x.rightBound());  
            if(right < left) 
               right = left;
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
          case 2:
            MpfrClass::roundUp();
            right = sin(x.rightBound());  
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
        } // gr_mod4 =  0:
        break;
      case 1:
        switch (ql_mod4) {
          case 0:
            MpfrClass::roundDown();
            left = sin(x.leftBound());
            right = sin(x.rightBound());  
            if(right < left) 
               left = right;
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
            
          case 1:
            MpfrClass::roundDown();
            left = sin(x.rightBound());
            MpfrClass::roundUp();
            right = sin(x.leftBound());  
            break;
            
          case 2:
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown); 
            break;
            
          case 3:
            MpfrClass::roundDown();
            left = sin(x.leftBound());
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
        }// gr_mod4 =  1:
        break;
      case 2:
        switch (ql_mod4) {
          case 0:
            MpfrClass::roundDown();
            left =  sin(x.rightBound());  
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
          
          case 1:
          case 2:
            MpfrClass::roundDown();
            left = sin(x.rightBound());
            MpfrClass::roundUp();
            right = sin(x.leftBound());  
            break;
          
          case 3:
            MpfrClass::roundDown();
            left = sin(x.leftBound());
            right = sin(x.rightBound());  
            if(right < left) 
               left = right;
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
        }// gr_mod4 =  2:
        break;
      case 3:
        switch (ql_mod4) {
          case 0:
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown); 
            break;
            
          case 1:
            MpfrClass::roundUp();
            right = sin(x.leftBound());  
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
            
          case 2:
            MpfrClass::roundUp();
            left = sin(x.leftBound());
            right = sin(x.rightBound());  
            if(right < left) 
               right = left;
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
            
          case 3:
            MpfrClass::roundDown();
            left = sin(x.leftBound());
            MpfrClass::roundUp();
            right = sin(x.rightBound());  
            break;
            
        } // gr_mod4 =  3:
        break;
      } // switch(gr_mod4)
  } 
  return Interval<MpfrClass>(left, right);
} // sin


template<> 
Interval<MpfrClass> cos(const Interval<MpfrClass>& x)
{
  Interval<MpfrClass>::BoundType quad_left, quad_right;
  Interval<MpfrClass>::BoundType left, right;

  int ql_mod4, qr_mod4;
 
  testNaN(x);

#ifdef _MPI_TEST_INF_   
  if (isinf(x)) 
  {
    // the two endpoints are the same infinite 
    if ( compare(x.leftBound(), x.rightBound()) == 0) 
       throw NaNException();
 
    return Interval<MpfrClass>(-1.0, 1.0);
  }
#endif //_MPI_TEST_INF_   
   
  // quad_left gives the quadrant where the left endpoint of b is 
  // quad_left = floor(2 b->left / pi)                            
  // idem for quad_right and b->right                             
  
  quad_left = quadrant(x.leftBound());
  quad_right = quadrant(x.rightBound());

  if (quad_right - quad_left >= 4.0) 
  {
     return Interval<MpfrClass> (-1, 1);
  } 
  else 
  {
  // there is less than one period in b 
  // let us discuss according to the position (quadrant) of the endpoints of b    
  
    
    
    // qr_mod4 gives the quadrant where the right endpoint of b is 
    // qr_mod4 = floor(2 b->right / pi) mod 4 
    qr_mod4 = modulo4(quad_right); // Warning: It do not work for huge numbers
  
    // quad_left gives the quadrant where the left endpoint of b is 
    // quad_left = floor(2 b->left / pi) mod 4 
    ql_mod4 = modulo4(quad_left);
              
    switch (qr_mod4) {
      case 0:
        switch (ql_mod4) {
           
          case 0:
            MpfrClass::roundDown();
            left = cos(x.rightBound());
            MpfrClass::roundUp();
            right = cos(x.leftBound());  
            break;
           
          case 1:
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown); 
            break;
            
          case 2:
            MpfrClass::roundDown();
            left = cos(x.leftBound());
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
          
          case 3:
            MpfrClass::roundDown();
            left = cos(x.leftBound());
            right = cos(x.rightBound());  
            if(right < left) 
               left = right;
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
          
        }
        break;
      case 1:
        switch (ql_mod4) {
          case 3:
            MpfrClass::roundDown();
            left =  cos(x.rightBound());  
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
         
          case 0:
          case 1:
            MpfrClass::roundDown();
            left = cos(x.rightBound());
            MpfrClass::roundUp();
            right = cos(x.leftBound());  
            break;
          
          case 2:
            MpfrClass::roundDown();
            left = cos(x.leftBound());
            right = cos(x.rightBound());  
            if(right < left) 
               left = right;
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            break;
        }
        break;
      case 2:
        switch (ql_mod4) {
            
          case 0:
            MpfrClass::roundUp();
            right = cos(x.leftBound());  
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
            
          case 1:
            MpfrClass::roundUp();
            left = cos(x.leftBound());
            right = cos(x.rightBound());  
            if(right < left) 
               right = left;
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
            
          case 2:
            MpfrClass::roundDown();
            left = cos(x.leftBound());
            MpfrClass::roundUp();
            right = cos(x.rightBound());  
            break;
        
          case 3:
            right = Interval<MpfrClass>::BoundType(1.0, capd::multiPrec::RoundUp);
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown); 
            break;
            
        }
        break;
        
        case 3:
        switch (ql_mod4) {
                    
          case 0:
            MpfrClass::roundUp();
            left = cos(x.leftBound());
            right = cos(x.rightBound());  
            if(right < left) 
               right = left;
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
            
          case 1:
            MpfrClass::roundUp();
            right = cos(x.rightBound());  
            left = Interval<MpfrClass>::BoundType(-1.0, capd::multiPrec::RoundDown);
            break;
            
          case 2:
          case 3:
            MpfrClass::roundDown();
            left = cos(x.leftBound());
            MpfrClass::roundUp();
            right = cos(x.rightBound());  
            break;
        }
        break;
      }
  }
  return Interval<MpfrClass>(left, right);
}  // cos



 
template<>
Interval<MpfrClass> tan(const Interval<MpfrClass>& x)
{
  testNaN(x);
  
#ifdef _MPI_TEST_INF_     
  if (isinf(x)) {
    /* the two endpoints are the same infinite */
    if ( compare(x.leftBound(), x.rightBound()) == 0) 
       throw NaNException();
   
    return Interval<MpfrClass>(Interval<MpfrClass>::BoundType::NegativeInfinity(), Interval<MpfrClass>::BoundType::PositiveInfinity());
  }
#endif //_MPI_TEST_INF_   

   
  Interval<MpfrClass>::BoundType quad_left = quadrant(x.leftBound()),
                                 quad_right = quadrant(x.rightBound());

  long int qr_mod2 = toInt(quad_right) % 2, // Warning: It do not work for huge numbers
           ql_mod2 = toInt(quad_left) % 2;
  
  /* if there is at least one period in b or if b contains a Pi/2 + k*Pi, */
  /* then a = ]-oo, +oo[ */ 
  if ((quad_right - quad_left >= 2.0)||((ql_mod2==0)&&(qr_mod2==1))) 
     return Interval<MpfrClass>(Interval<MpfrClass>::BoundType::NegativeInfinity(), Interval<MpfrClass>::BoundType::PositiveInfinity());

  /* within one period, tan is increasing */   
  Interval<MpfrClass>::BoundType left, right;
  MpfrClass::roundDown();
  left = tan(x.leftBound());
  MpfrClass::roundUp();
  right = tan(x.rightBound());
  return Interval<MpfrClass>(left, right);
}  //tan


template<>
Interval<MpfrClass> log(const Interval<MpfrClass>& x)
{
  Interval<MpfrClass>::BoundType left, right;
  MpfrClass::roundDown();
  left=log(x.leftBound());
  MpfrClass::roundUp();
  right=log(x.rightBound());
  
  Interval<MpfrClass> res(left, right);
  testNaN(res);
  
  return res;
}

template<>
Interval<MpfrClass> exp(const Interval<MpfrClass>& x)
{
  Interval<MpfrClass>::BoundType left, right;
  
  MpfrClass::roundDown();
  left=exp(x.leftBound());
  MpfrClass::roundUp();
  right=exp(x.rightBound());
  
  Interval<MpfrClass> res(left, right);
  testNaN(res);
  
  return res;
}

//-------- sqrt -----------------
template<>
Interval<MpfrClass> sqrt(const Interval<MpfrClass>  & z)
{
  Interval<MpfrClass>::BoundType  left, right;

  if(z.leftBound()<0.0){
     throw IntervalError<MpfrClass>("Interval<MpfrClass> Function sqrt(x) - domain error",z.leftBound(),z.rightBound());
  }
  MpfrClass::roundUp();
  right=sqrt(z.rightBound());

  MpfrClass::roundDown();
  left=sqrt(z.leftBound());

  return Interval<MpfrClass>(left, right);
}//-------------------------------


template<>
Interval<MpfrClass> power(const Interval<MpfrClass>& value, /*long*/  int exponent)
//   value^{exponent}
{
  Interval<MpfrClass>::BoundType l,r,t;
  
  int sign=1;
  
  if (exponent < 0 ) 
  {
    sign=0;
    exponent = -exponent;
  }

  l=value.leftBound();
  r=value.rightBound();

  /* if 'exponent' is even, then we need to correct the ends of interval -
     only absolute values matter*/
  if (exponent % 2 == 0)
  {
    l = abs(l);
    r = abs(r);
    if (l > r) 
    {
      t=r;       r=l;       l=t;
    }
    if (value.contains(0.0)) 
      l=0.0;
  }
  MpfrClass::roundUp();
  r = r.pow((long int)exponent); 

  MpfrClass::roundDown();
  l = l.pow((long int)exponent);

  Interval<MpfrClass> out(l, r);  

  if(sign==0)  // when an exponent is negative
  {
    out = Interval<MpfrClass>(1.0)/out;
  }

  return out;
}//--------------------------------------------------------------------

}}  // end of namespace capd::intervals

#endif  // __HAVE_MPFR__

#endif  // _MP_INTERVAL_FUN_

/// @}

