// This is TIMEUSED.H. Copyright (C) 2002-2003 by Pawel Pilarczyk.
// It contains a definition of a class used for measuring program's running
// time and displaying this time in a nice way.
// This is free software. No warranty. Consult 'license.txt' for details.
// You may want to include "textfile.h" before this file.
// Started on March 23, 2002. Last revision: February 9, 2003.

// NOTE: If you do not want to control the time measurement, but only want
// to make your program display a "Time used" message at its termination,
// then it is enough that you only link your program with "timeused.cpp"
// without even including this header file in your code.


#ifndef _TIMEUSED_H_
#define _TIMEUSED_H_

#include "options.h"

#include <ctime>
#include <iostream>


// classes defined within this header file:
class timeused;


//Doc//
// --------------------------------------------------
// ------------------- TIMEUSED ---------------------
// --------------------------------------------------

class timeused
{
public:
	// constructors
	timeused (const char *msg = NULL);
	timeused (std::ostream &out, const char *msg = NULL);

	// destructor
	~timeused ();

	// define an output stream for displaying program's
	// running time at program's termination
	timeused &operator = (std::ostream &out);

	#ifdef _TEXTFILE_H_
	// define an output stream for displaying program's
	// running time at program's termination to a pair of streams
	timeused &operator = (outputstream &out);
	#endif

	// turn off writing program's running time
	// at program's termination by assigning 0
	timeused &operator = (int n);

	// chang the message displayed at program's termination
	timeused &operator = (const char *msg);

	// reset the timer to the current moment
	timeused &reset ();

	// get the time from the initialization measured in seconds
	operator double ();

	// show time used from the beginning up to the current point;
	// the time is preceded with the message (default: Time used)
	void show (std::ostream &out, const char *message = NULL)
		const;
	void show (const char *message = NULL) const;

	// show time elapsed up to this point
	friend std::ostream &operator << (std::ostream &out,
		const timeused &t);

protected:
	// CPU usage start time (in seconds)
	double cpu0;
	
	// start time (in seconds)
	time_t t0;

	// output stream (NULL for no output)
	std::ostream *outstream1, *outstream2;

	// a message to display instead of "Used time" (if not NULL)
	const char *message;

	// should the constructor display the time?
	// note: -1 makes the destructor display times only > 1 sec
	int display;

}; /* timeused */
//End//

// --------------------------------------------------

inline timeused::timeused (const char *msg)
{
	reset ();
	outstream1 = NULL;
	outstream2 = NULL;
	message = msg;
	display = -1;
	return;
} /* timeused::timeused */

inline timeused::timeused (std::ostream &out, const char *msg)
{
	reset ();
	outstream1 = &out;
	outstream2 = NULL;
	message = msg;
	display = -1;
	return;
} /* timeused::timeused */

inline void timeused::show (std::ostream &out, const char *msg) const
{
	if (!msg)
		msg = message;
	if (!msg)
		msg = "Time used:";

	out << msg << ' ' << *this << '.' << std::endl;

	return;
} /* timeused::show */

inline timeused::~timeused ()
{
	if (!display || (!outstream1 && !outstream2))
		return;
	if ((display > 0) || (double (*this) > 1))
	{
		if (outstream1)
			show (*outstream1);
		if (outstream2)
			show (*outstream2);
	}
	return;
} /* timeused::~timeused */

inline timeused &timeused::operator = (std::ostream &out)
{
	outstream1 = &out;
	outstream2 = NULL;
	return *this;
} /* timeused::operator = */

#ifdef _TEXTFILE_H_
inline timeused &timeused::operator = (outputstream &out)
{
	if (out. show)
		outstream1 = &(out. out);
	else
		outstream1 = NULL;
	if (out. log)
		outstream2 = out. getlogstream ();
	else
		outstream2 = NULL;
	return *this;
} /* operator = */
#endif

inline timeused &timeused::operator = (int n)
{
	display = n;
	return *this;
} /* timeused::operator = */

inline timeused &timeused::operator = (const char *msg)
{
	message = msg;
	return *this;
} /* timeused::operator = */

inline void timeused::show (const char *msg) const
{
	if (outstream1)
		show (*outstream1, msg);
	if (outstream2)
		show (*outstream2, msg);
	return;
} /* timeused::show */

//Doc//
// --------------------------------------------------

extern timeused program_time;
//End//

#endif // _TIMEUSED_H_

